# Standard Library
import datetime
import logging
from typing import ClassVar

# Third Party
from eve_sde.models import ItemType, SolarSystem
from solo.models import SingletonModel

# Django
from django.db import models
from django.db.models import ExpressionWrapper, F, Func
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from .. import app_settings
from ..managers import AuditCharacterManager, AuditCorporationManager

logger = logging.getLogger(__name__)


class EveLocation(models.Model):
    location_id = models.BigIntegerField(primary_key=True)
    location_name = models.CharField(max_length=255)
    system = models.ForeignKey(
        SolarSystem,
        on_delete=models.SET_NULL,
        null=True,
        default=None
    )
    last_update = models.DateTimeField(auto_now=True)
    managed = models.BooleanField(default=False)
    managed_corp = models.ForeignKey(
        "CorporationAudit", default=None, blank=True, null=True, on_delete=models.CASCADE)
    managed_char = models.ForeignKey(
        "CharacterAudit", default=None, blank=True, null=True, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.location_name}"


class CorptoolsConfiguration(SingletonModel):
    holding_corps = models.ManyToManyField(EveCorporationInfo, blank=True)

    aggregate_lookback = models.IntegerField(
        default=90,
        help_text="Days to look back in aggregated views, like the mining ledger graphs."
    )

    disable_verification_assets = models.BooleanField(
        default=False,
        help_text="Allow ESI to provide data that does not match the ESI Assets Spec"
    )

    disable_verification_notifications = models.BooleanField(
        default=False,
        help_text="Allow ESI to provide data that does not match the ESI Notification Spec"
    )

    disable_update_pub_data = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Pulic History Data"
    )

    disable_update_skills = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character SKill Data"
    )

    disable_update_clones = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Clone Data"
    )

    disable_update_contacts = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Contact Data"
    )

    disable_update_contracts = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Contracts Data"
    )
    disable_update_assets = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Asset Data"
    )

    disable_update_wallet = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Wallet Data"
    )

    disable_update_notif = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Notification Data"
    )

    disable_update_roles = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Roles Data"
    )

    disable_update_mails = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Mail Data"
    )

    disable_update_location = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Location Data"
    )

    disable_update_loyaltypoints = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Loyalty Point Data"
    )

    disable_update_mining = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character Mining Data"
    )

    disable_update_indy = models.BooleanField(
        default=False,
        blank=True,
        help_text="Temporarily disable the ESI pulls for Character industry Job Data"
    )

    class Meta:
        permissions = (
            ('holding_corp_structures',
             'Can access configured holding corp structure data.'),
            ('holding_corp_wallets', 'Can access configured holding corp wallet data.'),
            ('holding_corp_assets', 'Can access configured holding corp asset data.')
        )

        default_permissions = []

    def __str__(self) -> str:
        return "Corptools Configuration"

    def holding_corp_qs(self):
        return CorporationAudit.objects.filter(corporation__in=self.holding_corps.all())

    @classmethod
    def skip_verify_assets(cls):
        try:
            return cls.get_solo().disable_verification_assets
        except Exception as e:
            logger.error(e)
            return True

    @classmethod
    def skip_verify_notifications(cls):
        try:

            return cls.get_solo().disable_verification_notifications
        except Exception as e:
            logger.error(e)
            return True


def check_date(last_update, time_ref):
    date = last_update if last_update is not None else timezone.now() - \
        datetime.timedelta(days=99)
    return date > time_ref


class CharacterAudit(models.Model):

    objects: ClassVar[AuditCharacterManager] = AuditCharacterManager()

    active = models.BooleanField(default=False)

    character = models.OneToOneField(EveCharacter, on_delete=models.CASCADE)

    last_update_pub_data = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_skills = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_skill_que = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_clones = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_contacts = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_contracts = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_assets = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_wallet = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_orders = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_notif = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_roles = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_titles = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_mails = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_location = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_loyaltypoints = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_mining = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_indy = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_login = models.DateTimeField(
        null=True, default=None, blank=True)

    balance = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None)

    last_known_login = models.DateTimeField(
        null=True, default=None, blank=True)

    last_known_logoff = models.DateTimeField(
        null=True, default=None, blank=True)

    total_logins = models.IntegerField(
        null=True, default=None, blank=True)

    def __str__(self):
        return f"{self.character.character_name}'s Character Data"

    class Meta:
        permissions = (('corp_hr', 'Can access other character\'s data for own corp.'),
                       ('alliance_hr',
                        'Can access other character\'s data for own alliance.'),
                       ('state_hr', 'Can access other character\'s data for own state.'),
                       ('guest_hr', 'Can access guest users\'s data for HR purposes.'),
                       ('global_hr', 'Can access other character\'s data for characters in any corp/alliance/state.'))

    def is_active(self):
        time_ref = timezone.now() - datetime.timedelta(days=app_settings.CT_CHAR_MAX_INACTIVE_DAYS)
        ct_conf = CorptoolsConfiguration.get_solo()

        try:
            is_active = True
            if app_settings.CT_CHAR_ACTIVE_IGNORE_CORP_HISTORY and not ct_conf.disable_update_pub_data:
                is_active = is_active and check_date(
                    self.last_update_pub_data, time_ref)
            if app_settings.CT_CHAR_ASSETS_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_ASSETS_MODULE and not ct_conf.disable_update_assets:
                is_active = is_active and check_date(
                    self.last_update_assets, time_ref)
            if app_settings.CT_CHAR_CLONES_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_CLONES_MODULE and not ct_conf.disable_update_clones:
                is_active = is_active and check_date(
                    self.last_update_clones, time_ref)
            if app_settings.CT_CHAR_SKILLS_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_SKILLS_MODULE and not ct_conf.disable_update_skills:
                is_active = is_active and check_date(
                    self.last_update_skills, time_ref)
                is_active = is_active and check_date(
                    self.last_update_skill_que, time_ref)
            if app_settings.CT_CHAR_WALLET_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_WALLET_MODULE and not ct_conf.disable_update_wallet:
                is_active = is_active and check_date(
                    self.last_update_wallet, time_ref)
                is_active = is_active and check_date(
                    self.last_update_orders, time_ref)
            if app_settings.CT_CHAR_NOTIFICATIONS_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_NOTIFICATIONS_MODULE and not ct_conf.disable_update_notif:
                is_active = is_active and check_date(
                    self.last_update_notif, time_ref)
            if app_settings.CT_CHAR_ROLES_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_ROLES_MODULE and not ct_conf.disable_update_roles:
                is_active = is_active and check_date(
                    self.last_update_roles, time_ref)
            if app_settings.CT_CHAR_MAIL_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_MAIL_MODULE and not ct_conf.disable_update_mails:
                is_active = is_active and check_date(
                    self.last_update_mails, time_ref)
            if app_settings.CT_CHAR_LOYALTYPOINTS_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_LOYALTYPOINTS_MODULE and not ct_conf.disable_update_loyaltypoints:
                is_active = is_active and check_date(
                    self.last_update_loyaltypoints, time_ref)
            if app_settings.CT_CHAR_MINING_MODULE and not app_settings.CT_CHAR_ACTIVE_IGNORE_MINING_MODULE and not ct_conf.disable_update_mining:
                is_active = is_active and check_date(
                    self.last_update_mining, time_ref)

            if self.active != is_active:
                self.active = is_active
                self.save()

            return is_active
        except Exception as e:
            logger.error(e, exc_info=True)
            return False

    @classmethod
    def get_oldest_qs(cls):
        time_ref = timezone.now() - datetime.timedelta(
            days=app_settings.CT_CHAR_MAX_INACTIVE_DAYS * 3
        )

        ct_conf = CorptoolsConfiguration.get_solo()
        qs = []

        if app_settings.CT_CHAR_ASSETS_MODULE and not (
            app_settings.CT_CHAR_ACTIVE_IGNORE_ASSETS_MODULE or ct_conf.disable_update_assets
        ):
            qs.append(Func(F('last_update_assets'), function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_CLONES_MODULE and not (
            app_settings.CT_CHAR_ACTIVE_IGNORE_CLONES_MODULE or ct_conf.disable_update_clones
        ):
            qs.append(Func(F('last_update_clones'), function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_SKILLS_MODULE and not (
           app_settings.CT_CHAR_ACTIVE_IGNORE_SKILLS_MODULE or ct_conf.disable_update_skills
        ):  # NOQA E124
            qs.append(Func(F('last_update_skills'), function='UNIX_TIMESTAMP'))
            qs.append(Func(F('last_update_skill_que'),
                      function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_WALLET_MODULE and not (
           app_settings.CT_CHAR_ACTIVE_IGNORE_WALLET_MODULE or ct_conf.disable_update_wallet
        ):  # NOQA E124
            qs.append(Func(F('last_update_wallet'), function='UNIX_TIMESTAMP'))
            qs.append(Func(F('last_update_orders'), function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_NOTIFICATIONS_MODULE and not (
           app_settings.CT_CHAR_ACTIVE_IGNORE_NOTIFICATIONS_MODULE or ct_conf.disable_update_notif
        ):  # NOQA E124
            qs.append(Func(F('last_update_notif'), function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_ROLES_MODULE and not (
            app_settings.CT_CHAR_ACTIVE_IGNORE_ROLES_MODULE or ct_conf.disable_update_roles
        ):
            qs.append(Func(F('last_update_roles'), function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_MAIL_MODULE and not (
            app_settings.CT_CHAR_ACTIVE_IGNORE_MAIL_MODULE or ct_conf.disable_update_mails
        ):
            qs.append(Func(F('last_update_mails'), function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_LOYALTYPOINTS_MODULE and not (
            app_settings.CT_CHAR_ACTIVE_IGNORE_LOYALTYPOINTS_MODULE or ct_conf.disable_update_loyaltypoints
        ):
            qs.append(Func(F('last_update_loyaltypoints'),
                      function='UNIX_TIMESTAMP'))

        if app_settings.CT_CHAR_MINING_MODULE and not (
            app_settings.CT_CHAR_ACTIVE_IGNORE_MINING_MODULE or ct_conf.disable_update_mining
        ):
            qs.append(Func(F('last_update_mining'), function='UNIX_TIMESTAMP'))

        tot = len(qs)

        qout = qs.pop()

        for q in qs:
            qout = qout + q

        query = cls.objects.annotate(
            avg_date=Func(
                ExpressionWrapper(
                    qout / tot,
                    output_field=models.BigIntegerField()
                ),
                function='FROM_UNIXTIME',
                output_field=models.DateTimeField()
            )
        ).filter(
            character__character_ownership__isnull=False,
            avg_date__gte=time_ref
        ).order_by("avg_date")

        return query


class CorporationAudit(models.Model):

    objects: ClassVar[AuditCorporationManager] = AuditCorporationManager()

    corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE)

    last_update_pub_data = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_pub_data = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_assets = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_assets = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_structures = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_structures = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_moons = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_moons = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_observers = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_observers = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_wallet = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_wallet = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_contracts = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_contracts = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_industry_jobs = models.DateTimeField(
        null=True, default=None, blank=True)
    last_change_industry_jobs = models.DateTimeField(
        null=True, default=None, blank=True)

    last_update_known_login = models.DateTimeField(
        null=True, default=None, blank=True)

    def __str__(self):
        return f"{self.corporation.corporation_name}'s Corporation Data"

    class Meta:
        permissions = (
            ('own_corp_manager',
             'Can access own corporations\'s data.'),
            ('alliance_corp_manager',
             'Can access other corporations\'s data for own alliance.'),
            ('state_corp_manager',
             'Can access other corporations\'s data for own state.'),
            ('global_corp_manager', 'Can access all corporations\'s data.'),
            ('show_if_director',
             'Can access all corporations\'s where character is a director.'),
        )


class CharacterLocation(models.Model):
    character = models.OneToOneField(
        CharacterAudit, on_delete=models.CASCADE, related_name="location")

    current_location = models.ForeignKey(
        EveLocation,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        default=None
    )
    current_ship = models.ForeignKey(
        ItemType,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        default=None
    )
    current_ship_unique = models.BigIntegerField(
        null=True,
        blank=True,
        default=None
    )
    current_ship_name = models.TextField(
        max_length=150,
        null=True,
        blank=True,
        default=None
    )



# Third Party
from eve_sde.models import (
    SolarSystem,
)

# Django
from django.core.cache import cache

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from esi.exceptions import HTTPClientError, HTTPNotModified
from esi.models import Token

# AA Example App
from corptools import providers
from corptools.models import EveLocation

logger = get_extension_logger(__name__)


def set_error_count_flag():
    return cache.set("esi_errors_timeout", 1, 60)


def fetch_location_name(location_id, location_flag, character_id, update=False):
    """Takes a location_id and character_id and returns a location model for items in a station/structure or in space"""

    accepted_location_flags = ['AssetSafety',
                               'Deliveries',
                               'Hangar',
                               'HangarAll',
                               'solar_system',
                               'OfficeFolder',
                               'AutoFit']

    if location_flag not in accepted_location_flags:
        if location_flag is not None:
            return None  # ship fits or in cargo holds or what ever also dont care

    existing = EveLocation.objects.filter(location_id=location_id)
    current_loc = existing.exists()

    if current_loc and location_id < 64000000:
        return existing.first()
    else:
        existing = existing.first()

    if location_id == 2004:
        # ASSET SAFETY
        return EveLocation(location_id=location_id,
                           location_name="Asset Safety")
    elif 30000000 < location_id < 33000000:  # Solar System
        system = SolarSystem.objects.filter(id=location_id)
        if not system.exists():
            logger.error("Unknown System, Have you populated the map?")
            # TODO Do i fire the map population task here?
            return None
        else:
            system = system.first()
        return EveLocation(location_id=location_id,
                           location_name=system.name,
                           system=system)
    elif 60000000 < location_id < 64000000:  # Station ID
        station = providers.esi_openapi.client.Universe.GetUniverseStationsStationId(
            station_id=location_id
        ).result()
        system = SolarSystem.objects.filter(id=station.system_id)
        if not system.exists():
            logger.error("Unknown System, Have you populated the map?")
            # TODO Do i fire the map population task here?
            return None
        return EveLocation(
            location_id=location_id,
            location_name=station.name,
            system_id=station.system_id
        )

    req_scopes = ['esi-universe.read_structures.v1']

    token = Token.get_token(character_id, req_scopes)

    if not token:
        return None

    else:  # Structure id?
        try:
            structure = providers.esi_openapi.client.Universe.GetUniverseStructuresStructureId(
                structure_id=location_id,
                token=token
            ).result(use_etag=False)
        except HTTPClientError as e:  # no access
            if int(e.headers.get('x-esi-error-limit-remain')) < 50:
                set_error_count_flag()
            logger.debug(
                "Failed to get location:{}, Error:{}, Errors Remaining:{}, Time Remaining: {}".format(
                    location_id,
                    e,
                    e.headers.get('x-esi-error-limit-remain'),
                    e.headers.get('x-esi-error-limit-reset')
                )
            )
            return None
        system = SolarSystem.objects.filter(
            id=structure.solar_system_id
        )
        if not system.exists():
            logger.error("Unknown System, Have you populated the map?")
            # TODO Do i fire the map population task here?
            return None
        if current_loc:
            existing.location_name = structure.name
            return existing
        else:
            return EveLocation(location_id=location_id,
                               location_name=structure.name,
                               system_id=structure.solar_system_id)



# Generated by Django 4.2.16 on 2026-02-27 11:03
# Standard Library
from datetime import timedelta

# Django
import django.db.models.deletion
from django.db import migrations, models
from django.utils import timezone


def clear(qryset, model_name):
    """ Tidy up the old unused models. """
    if qryset.exists():
        for o in qryset:
            print(f" - Deleting {model_name} record {vars(o)}")
            o.delete()


def check(qryset, model_name):
    """ Check for any records that would be left orphaned by the migration. """
    count = qryset.count()
    if count > 0:
        items = "\n".join([f" - {vars(o)})" for o in qryset])
        raise Exception(
            f"There are {count} {model_name} records that do not have a corresponding record in the eve_sde app. \n"
            "Please run the esde_load_sde management command to sync the data before applying this migration. \n"
            f"{items}"
        )


def check_exists(apps, schema_editor):
    es_update = apps.get_model('eve_sde', 'EveSDE')

    last_sde_update = es_update.objects.all().first().last_check_date

    print(" Running pre-migration check for eve_sde data freshness...")
    if last_sde_update is None or last_sde_update < timezone.now() - timedelta(days=1):
        print(last_sde_update)
        raise Exception(
            "The eve_sde app has not been updated in the last 24h.\n"
            "Please run the `esde_load_sde` management command to sync the data before applying this migration.")
    print(" eve_sde data is up to date. Proceeding with migration...")

    ct_item = apps.get_model('corptools', 'EveItemType')
    es_item = apps.get_model('eve_sde', 'ItemType')
    clear(ct_item.objects.exclude(
        type_id__in=es_item.objects.all().values_list("id", flat=True)), "EveItemType")
    check(ct_item.objects.exclude(
        type_id__in=es_item.objects.all().values_list("id", flat=True)), "EveItemType")

    ct_grp = apps.get_model('corptools', 'EveItemGroup')
    es_grp = apps.get_model('eve_sde', 'ItemGroup')
    clear(ct_grp.objects.exclude(group_id__in=es_grp.objects.all(
    ).values_list("id", flat=True)), "EveItemGroup")
    check(ct_grp.objects.exclude(group_id__in=es_grp.objects.all(
    ).values_list("id", flat=True)), "EveItemGroup")

    ct_cat = apps.get_model('corptools', 'EveItemCategory')
    es_cat = apps.get_model('eve_sde', 'ItemCategory')
    clear(ct_cat.objects.exclude(category_id__in=es_cat.objects.all(
    ).values_list("id", flat=True)), "EveItemCategory")
    check(ct_cat.objects.exclude(category_id__in=es_cat.objects.all(
    ).values_list("id", flat=True)), "EveItemCategory")

    ct_reg = apps.get_model('corptools', 'MapRegion')
    es_reg = apps.get_model('eve_sde', 'Region')
    clear(ct_reg.objects.exclude(
        region_id__in=es_reg.objects.all().values_list("id", flat=True)), "MapRegion")
    check(ct_reg.objects.exclude(
        region_id__in=es_reg.objects.all().values_list("id", flat=True)), "MapRegion")

    ct_con = apps.get_model('corptools', 'MapConstellation')
    es_con = apps.get_model('eve_sde', 'Constellation')
    clear(ct_con.objects.exclude(constellation_id__in=es_con.objects.all(
    ).values_list("id", flat=True)), "MapConstellation")
    check(ct_con.objects.exclude(constellation_id__in=es_con.objects.all(
    ).values_list("id", flat=True)), "MapConstellation")

    ct_sys = apps.get_model('corptools', 'MapSystem')
    es_sys = apps.get_model('eve_sde', 'SolarSystem')
    clear(ct_sys.objects.exclude(
        system_id__in=es_sys.objects.all().values_list("id", flat=True)), "MapSystem")
    check(ct_sys.objects.exclude(
        system_id__in=es_sys.objects.all().values_list("id", flat=True)), "MapSystem")

    ct_pnt = apps.get_model('corptools', 'MapSystemPlanet')
    es_pnt = apps.get_model('eve_sde', 'Planet')
    clear(ct_pnt.objects.exclude(planet_id__in=es_pnt.objects.all(
    ).values_list("id", flat=True)), "MapSystemPlanet")
    check(ct_pnt.objects.exclude(planet_id__in=es_pnt.objects.all(
    ).values_list("id", flat=True)), "MapSystemPlanet")

    ct_mns = apps.get_model('corptools', 'MapSystemMoon')
    es_mns = apps.get_model('eve_sde', 'Moon')
    clear(ct_mns.objects.exclude(moon_id__in=es_mns.objects.all(
    ).values_list("id", flat=True)), "MapSystemMoon")
    check(ct_mns.objects.exclude(moon_id__in=es_mns.objects.all(
    ).values_list("id", flat=True)), "MapSystemMoon")


class Migration(migrations.Migration):
    dependencies = [
        ("eve_sde", "0012_alter_constellation_region"),
        ("corptools", "0129_alter_notification_id"),
    ]

    operations = [
        migrations.RunPython(check_exists),
        migrations.AlterField(
            model_name="assetsfilter",
            name="categories",
            field=models.ManyToManyField(
                blank=True,
                help_text="Filter on Asset Categories.",
                to="eve_sde.itemcategory",
            ),
        ),
        migrations.AlterField(
            model_name="assetsfilter",
            name="constellations",
            field=models.ManyToManyField(
                blank=True,
                help_text="Limit filter to specific constellations",
                to="eve_sde.constellation",
            ),
        ),
        migrations.AlterField(
            model_name="assetsfilter",
            name="groups",
            field=models.ManyToManyField(
                blank=True, help_text="Filter on Asset Groups.", to="eve_sde.itemgroup"
            ),
        ),
        migrations.AlterField(
            model_name="assetsfilter",
            name="regions",
            field=models.ManyToManyField(
                blank=True,
                help_text="Limit filter to specific regions",
                to="eve_sde.region",
            ),
        ),
        migrations.AlterField(
            model_name="assetsfilter",
            name="systems",
            field=models.ManyToManyField(
                blank=True,
                help_text="Limit filter to specific systems",
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="assetsfilter",
            name="types",
            field=models.ManyToManyField(
                blank=True, help_text="Filter on Asset Types.", to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="characterasset",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="characterbountystat",
            name="type_name",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="characterindustryjob",
            name="blueprint_type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="blueprint_type",
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="characterindustryjob",
            name="product_type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="product_type",
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="characterlocation",
            name="current_ship",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="charactermarketorder",
            name="region_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.region",
            ),
        ),
        migrations.AlterField(
            model_name="charactermarketorder",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="characterminingledger",
            name="system",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="eve_sde.solarsystem"
            ),
        ),
        migrations.AlterField(
            model_name="characterminingledger",
            name="type_name",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="contractitem",
            name="type_name",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="corpasset",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="corporatecontractitem",
            name="type_name",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="corporationindustryjob",
            name="blueprint_type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="corp_blueprint_type",
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="corporationindustryjob",
            name="product_type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="corp_product_type",
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="corporationmarketorder",
            name="region_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.region",
            ),
        ),
        migrations.AlterField(
            model_name="corporationmarketorder",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="currentshipfilter",
            name="constellations",
            field=models.ManyToManyField(
                blank=True,
                help_text="Limit filter to specific constellations",
                to="eve_sde.constellation",
            ),
        ),
        migrations.AlterField(
            model_name="currentshipfilter",
            name="groups",
            field=models.ManyToManyField(
                blank=True, help_text="Filter on Ship Groups.", to="eve_sde.itemgroup"
            ),
        ),
        migrations.AlterField(
            model_name="currentshipfilter",
            name="regions",
            field=models.ManyToManyField(
                blank=True,
                help_text="Limit filter to specific regions",
                to="eve_sde.region",
            ),
        ),
        migrations.AlterField(
            model_name="currentshipfilter",
            name="systems",
            field=models.ManyToManyField(
                blank=True,
                help_text="Limit filter to specific systems",
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="currentshipfilter",
            name="types",
            field=models.ManyToManyField(
                blank=True, help_text="Filter on Ship Types.", to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="evelocation",
            name="system",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="implant",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="mapjumpbridge",
            name="from_solar_system",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="bridge_from_system",
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="mapjumpbridge",
            name="to_solar_system",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="bridge_to_system",
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="oretax",
            name="item",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.DO_NOTHING, to="eve_sde.itemtype"
            ),
        ),
        migrations.AlterField(
            model_name="poco",
            name="planet",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.planet",
            ),
        ),
        migrations.AlterField(
            model_name="poco",
            name="system_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="skill",
            name="skill_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="skillqueue",
            name="skill_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="starbase",
            name="moon",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eve_sde.moon",
            ),
        ),
        migrations.AlterField(
            model_name="starbase",
            name="system",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="starbase",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="system_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="type_name",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="typeprice",
            name="item",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.DO_NOTHING, to="eve_sde.itemtype"
            ),
        ),
    ]



# Standard Library
import datetime
from typing import ClassVar

# Third Party
# Map Shims
# Item Shims
from eve_sde.models import ItemType, SolarSystem

# Django
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver

# AA Example App
from corptools.models.utils import JSONModel

from ..managers import (
    EveCategoryManager,
    EveGroupManager,
    EveItemTypeManager,
    EveMoonManager,
    EveNameManager,
    EvePlanetManager,
)

# EVE SDE Migration Start


class EveName(models.Model):
    objects: ClassVar[EveNameManager] = EveNameManager()

    eve_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    category = models.CharField(max_length=50)

    # optionals for character/corp
    corporation = models.ForeignKey(
        'EveName',
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="corp"
    )
    alliance = models.ForeignKey(
        'EveName',
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="alli"
    )
    last_update = models.DateTimeField(auto_now=True)

    CHARACTER = "character"
    CORPORATION = "corporation"
    ALLIANCE = "alliance"

    def __str__(self):
        return self.name

    def get_image_url(self):
        if self.category == self.CHARACTER:
            return eveimageserver.character_portrait_url(self.eve_id)
        elif self.category == self.CORPORATION:
            return eveimageserver.corporation_logo_url(self.eve_id)
        elif self.category == self.ALLIANCE:
            return eveimageserver.alliance_logo_url(self.eve_id)
        elif self.category == 'faction':  # CCP...
            return eveimageserver.corporation_logo_url(self.eve_id)

    def needs_update(self):
        return self.last_update + datetime.timedelta(days=30) < timezone.now()


class MapJumpBridge(models.Model):
    structure_id = models.BigIntegerField(primary_key=True)
    from_solar_system = models.ForeignKey(
        SolarSystem,
        on_delete=models.CASCADE,
        related_name="bridge_from_system"
    )
    to_solar_system = models.ForeignKey(
        SolarSystem,
        on_delete=models.CASCADE,
        related_name="bridge_to_system"
    )
    owner = models.ForeignKey(
        EveName, on_delete=models.SET_NULL, null=True, default=None)
    updated = models.DateTimeField(auto_now=True)
    manually_input = models.BooleanField(default=False)

    def __str__(self):
        return f"{self.from_solar_system.name} >> {self.to_solar_system.name} ({self.structure_id}) (Auto: {not self.manually_input})"


class TypePrice(models.Model):
    item = models.ForeignKey(ItemType, on_delete=models.DO_NOTHING)
    price = models.DecimalField(max_digits=20, decimal_places=2)
    last_update = models.DateTimeField(auto_now=True)


# ************ is this needed any more?

class OreTax(models.Model):
    item = models.ForeignKey(ItemType, on_delete=models.DO_NOTHING)
    price = models.DecimalField(max_digits=20, decimal_places=2)
    last_update = models.DateTimeField(auto_now=True)


class OreTaxRates(models.Model):
    tag = models.CharField(max_length=500, default="Mining Tax")
    refine_rate = models.DecimalField(
        max_digits=5, decimal_places=2, default=87.5)
    ore_rate = models.DecimalField(max_digits=5, decimal_places=2)  # normal
    ubiquitous_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # ubiq
    common_rate = models.DecimalField(max_digits=5, decimal_places=2)  # comon
    uncommon_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # uncom
    rare_rate = models.DecimalField(max_digits=5, decimal_places=2)  # rare
    excptional_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # best

# DEPRECATED MODELS BELOW - NOT USED ANYMORE
# These will be removed in a future release expected 4.0.0
# Use eve_sde instead https://github.com/Solar-Helix-Independent-Transport/django-eveonline-sde/tree/main


class EveItemCategory(JSONModel):
    """
    categories.jsonl
    """
    objects: ClassVar[EveCategoryManager] = EveCategoryManager()
    category_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)  # unknown max

    def __str__(self):
        return self.name


class EveItemGroup(JSONModel):
    """
    groups.jsonl
    """
    objects: ClassVar[EveGroupManager] = EveGroupManager()
    group_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)  # unknown max
    category = models.ForeignKey(
        EveItemCategory, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return self.name


class EveItemType(JSONModel):
    """
    types.jsonl
    """
    objects: ClassVar[EveItemTypeManager] = EveItemTypeManager()
    type_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)  # unknown max
    group = models.ForeignKey(
        EveItemGroup, on_delete=models.SET_NULL, null=True, default=None)
    description = models.TextField(null=True, default=None)
    mass = models.FloatField(null=True, default=None)
    packaged_volume = models.FloatField(null=True, default=None)
    portion_size = models.FloatField(null=True, default=None)
    volume = models.FloatField(null=True, default=None)
    published = models.BooleanField()
    radius = models.FloatField(null=True, default=None)

    def __str__(self):
        return self.name


class EveItemDogmaAttribute(JSONModel):
    """
    typeDogma.jsonl
    """
    eve_type = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None)
    attribute_id = models.BigIntegerField(null=True, default=None)
    value = models.FloatField(null=True, default=None)


class InvTypeMaterials(JSONModel):
    """
    typeDogma.jsonl
    """
    eve_type = models.ForeignKey(
        EveItemType,
        on_delete=models.SET_NULL,
        null=True,
        default=None
    )
    qty = models.IntegerField()
    type_id = models.IntegerField()
    material_type_id = models.IntegerField()
    met_type = models.ForeignKey(
        EveItemType,
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="met_type"
    )


class MapRegion(JSONModel):
    """
    mapRegions.jsonl
    """
    _filename = "mapRegions.jsonl"
    _update_fields = [
        "name",
        "description"
    ]

    region_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, default=None, )

    def __str__(self):
        return self.name


class MapConstellation(JSONModel):
    """
    mapConstellations.jsonl
    """
    constellation_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    region = models.ForeignKey(
        MapRegion, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return self.name


class MapSystem(JSONModel):
    """
    mapSolarSystems.jsonl
    """
    system_id = models.BigIntegerField(primary_key=True)
    security_status = models.FloatField()
    name = models.CharField(max_length=255)
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()
    security_class = models.CharField(max_length=255, null=True, default=None)
    star_id = models.IntegerField(null=True, default=None)
    constellation = models.ForeignKey(
        MapConstellation, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return self.name


class MapSystemGate(JSONModel):
    """
    mapStargates.jsonl
    """
    from_solar_system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="from_system")
    to_solar_system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="to_system")

    def __str__(self):
        return (self.from_solar_system_id, self.to_solar_system_id)


class MapSystemPlanet(JSONModel):
    """
    mapPlanets.jsonl
        "system_name planet_roman_numeral"
    """

    objects: ClassVar[EvePlanetManager] = EvePlanetManager()

    planet_id = models.IntegerField(primary_key=True)
    system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="planet")
    name = models.CharField(max_length=255)

    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()

    eve_type = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return (self.name)

    @classmethod
    def name_lookup(cls):
        return {
            s.get("system_id"): s.get("name") for s in
            MapSystem.objects.all().values("system_id", "name")
        }


class MapSystemMoon(JSONModel):
    """
    mapMoons.jsonl
        "system_name planet_roman_numeral - Moon #"
    """
    objects: ClassVar[EveMoonManager] = EveMoonManager()

    moon_id = models.IntegerField(primary_key=True)
    system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="moon")
    name = models.CharField(max_length=255)

    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()

    def __str__(self):
        return (self.name)

    @classmethod
    def name_lookup(cls):
        return {
            p.get("planet_id"): p.get("name") for p in
            MapSystemPlanet.objects.all().values("planet_id", "name")
        }



# Standard Library
import math
from typing import List

# Third Party
from eve_sde.models import TypeDogma
from ninja import NinjaAPI

# Django
from django.utils import timezone
from django.utils.translation import gettext as _

# AA Example App
from corptools import models, providers
from corptools.api import schema
from corptools.api.helpers import get_alts_queryset, get_main_character


class SkillApiEndpoints:

    tags = ["Account"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "account/{character_id}/skills",
            response={200: List[schema.CharacterSkills], 403: str},
            tags=self.tags
        )
        def get_character_skills(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skills = models.Skill.objects.filter(character__character__in=characters)\
                .select_related('character__character', 'skill_name', "skill_name__group")

            totals = models.SkillTotals.objects.filter(character__character__in=characters)\
                .select_related('character__character')

            output = {}

            for s in skills:
                if s.character_id not in output:
                    output[s.character_id] = {
                        "character": s.character.character,
                        "skills": [],
                        "total_sp": 0,
                        "unallocated_sp": 0
                    }
                output[s.character_id]["skills"].append(
                    {
                        "group": s.skill_name.group.name,
                        "skill": s.skill_name.name,
                        "sp": s.skillpoints_in_skill,
                        "level": s.trained_skill_level,
                        "active": s.active_skill_level,
                    }
                )

            for t in totals:
                if t.character_id in output:
                    output[t.character_id]["unallocated_sp"] = t.unallocated_sp
                    output[t.character_id]["total_sp"] = t.total_sp

            return list(output.values())

        @api.get(
            "account/{character_id}/skillgraph",
            response={200: dict, 403: str},
            tags=self.tags
        )
        def get_character_skillgraph(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            # characters = get_alts_queryset(main)
            skills = models.Skill.objects.filter(
                character__character__character_id=character_id
                # character__character__character_id__in=list(
                #     characters.order_by(
                #         "-characteraudit__skilltotals__total_sp"
                #     ).values_list("character_id", flat=True)[:10]
                # )
            ).select_related(
                'character__character',
                'skill_name',
                "skill_name__group"
            )

            dogmas_qry = TypeDogma.objects.filter(
                item_type__group__category_id=16,
                dogma_attribute_id=275
            ).select_related("item_type__group")

            character_skills = {}
            skill_dogma = {}

            for d in dogmas_qry:
                sp_5 = 250 * d.value * math.sqrt(32) ** 4
                skill_dogma[d.item_type.name] = sp_5
                if d.item_type.group.name not in character_skills:
                    character_skills[d.item_type.group.name] = {
                        "group": d.item_type.group.name,
                        "char_total": 0,
                        "total": 0
                    }
                character_skills[d.item_type.group.name]["total"] += sp_5

            character_names = set()

            for s in skills:
                character_names.add(s.character.character.character_name)

                if s.character.character.character_name not in character_skills[s.skill_name.group.name]:
                    character_skills[s.skill_name.group.name][s.character.character.character_name] = 0

                if s.skillpoints_in_skill > 0:
                    ratio = skill_dogma[s.skill_name.name] / \
                        character_skills[s.skill_name.group.name]["total"]
                    value = (s.skillpoints_in_skill /
                             skill_dogma[s.skill_name.name] * 100)

                    character_skills[s.skill_name.group.name][s.character.character.character_name] += value * ratio
                    character_skills[s.skill_name.group.name]["char_total"] += s.skillpoints_in_skill

            output = []
            for k, d in dict(sorted(character_skills.items())).items():
                tot = 0
                for c in character_names:
                    if c not in d:
                        d[c] = 0
                    else:
                        d[c] = round(d[c])
                        tot = tot if tot > d[c] else d[c]
                if tot > 5:
                    output.append(d)

            return {"characters": list(character_names), "data": output}

        @api.get(
            "account/{character_id}/skill/history",
            response={200: List[schema.CharacterSkillHistory], 403: str},
            tags=self.tags
        )
        def get_character_skill_history(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skill_history = models.SkillTotalHistory.objects.filter(character__character__in=characters)\
                .select_related('character__character').order_by("date")

            output = {}

            for s in skill_history:
                if s.character_id not in output:
                    output[s.character_id] = {
                        "character": s.character.character,
                        "history": [],
                    }
                output[s.character_id]["history"].append(
                    {
                        "date": s.date,
                        "sp": s.sp,
                        "total_sp": s.total_sp,
                        "unallocated_sp": s.unallocated_sp,
                    }
                )

            return list(output.values())

        @api.get(
            "account/{character_id}/skillqueues",
            response={200: List[schema.CharacterQueue], 403: str},
            tags=self.tags
        )
        def get_character_skillqueues(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            now = timezone.now()

            skills = models.SkillQueue.objects.filter(
                character__character__in=characters,
            ).select_related(
                "character__character",
                "skill_name",
                "skill_name__group"
            )

            skill_totals = models.Skill.objects.filter(
                character__character__in=characters
            ).select_related(
                "character__character",
                "skill_name",
                "skill_name__group"
            )

            skl_ttl = {}

            for s in skill_totals:
                if s.character_id not in skl_ttl:
                    skl_ttl[s.character_id] = {}
                skl_ttl[s.character_id][s.skill_name.name] = s.trained_skill_level

            output = {}
            for c in characters:
                output[c.character_id] = {
                    "character": c,
                    "queue": [],
                }

            for s in skills:
                if s.character.character.character_id not in output:
                    output[s.character.character.character_id] = {
                        "character": s.character.character,
                        "queue": [],
                    }

                if s.finish_date:
                    if s.finish_date < now:
                        continue

                output[s.character.character.character_id]["queue"].append(
                    {
                        "position": s.queue_position,
                        "group": s.skill_name.group.name,
                        "skill": s.skill_name.name,
                        "end_level": s.finish_level,
                        "start_sp": s.level_start_sp,
                        "end_sp": s.level_end_sp,
                        "start": s.start_date,
                        "end": s.finish_date,
                        "current_level": skl_ttl.get(
                            s.character.character_id,
                            {}
                        ).get(
                            s.skill_name.name,
                            0
                        )
                    }
                )

            return list(output.values())

        @api.get(
            "account/{character_id}/doctrines",
            response={200: List[schema.CharacterDoctrines], 403: str},
            tags=self.tags
        )
        def get_character_doctrines(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            skilllists = providers.skills.get_and_cache_user(
                main.character_ownership.user_id)
            visibles = list(models.SkillList.objects.filter(
                show_on_audit=True).values_list("name", flat=True))
            output = {}
            for c in characters:
                output[c.character_id] = {
                    "character": c,
                    "doctrines": {},
                    "skills": {},
                }

            for k, s in skilllists['skills_list'].items():
                for k, d in s["doctrines"].items():
                    # filter out hidden items
                    if k in visibles:
                        output[s['character_id']]["doctrines"][k] = d
                output[s['character_id']]["skills"] = s["skills"]

            return list(output.values())



# Standard Library
import datetime
from functools import wraps

# Third Party
from celery import signature
from celery.exceptions import Retry
from celery_once import AlreadyQueued
from httpx import RequestError

# Django
from django.core.cache import cache
from django.db.models import QuerySet
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from esi.exceptions import (
    ESIBucketLimitException,
    HTTPClientError,
    HTTPServerError,
)
from esi.rate_limiting import interval_to_seconds

logger = get_extension_logger(__name__)


def enqueue_next_task(chain, delay=1):
    """
        Queue next task, and attach the rest of the chain to it.
    """
    while (len(chain)):
        _t = chain.pop(0)
        _t = signature(_t)
        _t.kwargs.update({"chain": chain})
        try:
            _t.apply_async(priority=6, countdown=delay)
        except AlreadyQueued:
            # skip this task as it is already in the queue
            logger.warning(f"Skipping task as its already queued {_t}")
            continue
        break


def set_error_flag(timeout):
    tout = timezone.now() + datetime.timedelta(seconds=timeout)
    cache.set("esi_error_timeout", tout, timeout=timeout + 1)


def get_error_flag():
    return cache.get("esi_error_timeout", default=timezone.now())


def clear_error_flag():
    cache.delete("esi_error_timeout")


def esi_error_retry(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        _ret = None

        if get_error_flag() >= timezone.now():
            logger.warning("Hit ESI error limit! will retry tasks!")
            args[0].retry(countdown=61)
        else:
            clear_error_flag()
        try:
            _ret = func(*args, **kwargs)
        except Exception as e:
            if isinstance(e, (HTTPClientError, HTTPServerError)):  # Bravado, OpenAPI
                code = e.status_code
                if code in (420, 429):
                    logger.warning(f"Hit ESI error limit! Pausing Tasks! {e}")
                    set_error_flag(60)
                    args[0].retry(countdown=61)
            elif isinstance(e, (ESIBucketLimitException)):  # OpenAPI
                logger.warning(f"Hit ESI rate bucket! Pausing Task! {e}")
                args[0].retry(countdown=e.reset)
            elif isinstance(e, RequestError):
                logger.warning(f"Uncaught Error from HTTPX, Retrying... {e}")
                args[0].retry(countdown=30)
            elif isinstance(e, (OSError)):  # Bravado
                logger.warning(f"Hit ESI error! Skipping task! {e}")
            raise e
        return _ret
    return wrapper


def no_fail_chain(func):
    """
        Decorator to chain tasks provided in the chain kwargs regardless of task failures.
        Be sure to add chain=[] to your kwargs. TODO make this not needed.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        excp = None
        _ret = None
        try:
            _ret = func(*args, **kwargs)
        except Exception as e:
            excp = e
        finally:
            _chn = kwargs.get("chain", [])
            if not isinstance(excp, Retry):
                enqueue_next_task(_chn)
            if excp:
                raise excp
        return _ret
    return wrapper


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    total = 0
    if isinstance(lst, list):
        total = len(lst)
    elif isinstance(lst, QuerySet):
        total = lst.count()

    for i in range(0, total, n):
        yield lst[i:i + n]


def limit_to_rate(rate_limit):
    """convert things like 100/15m or 10/s to a delay in seconds"""
    lim = rate_limit.split("/")
    if len(lim[1]) < 2:
        secs = interval_to_seconds(f"1{lim[1]}")
    else:
        secs = interval_to_seconds(lim[1])
    return secs/int(lim[0])


def rate_limited_task_no_fail(rate_limit):
    delay = limit_to_rate(rate_limit)

    def decorator(func):
        def rate_lim_wrapper(*args, **kwargs):
            if args[0].request.retries == 0:
                # If it's our first go we will wait for the rate limit time
                args[0].retry(countdown=delay)
            excp = None
            try:
                return func(*args, **kwargs)
            except Exception as e:
                excp = e
                logger.exception(e)
            finally:
                if isinstance(excp, Retry):
                    raise excp
        return rate_lim_wrapper
    return decorator



# Standard Library
from datetime import timedelta

# Third Party
from ninja import Field, Schema
from ninja.pagination import LimitOffsetPagination
from ninja.types import DictStrAny

# Django
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F, Q, QuerySet, Sum
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# AA Example App
from corptools.constants.types import (
    MINING_GAS_GROUPS,
    MINING_ICE_GROUPS,
    MINING_MOON_GROUPS,
    MINING_ORE_GROUPS,
    RAT_CAPITAL_GROUPS,
    RAT_OFFICER_CRUISER_GROUPS,
    RAT_OFFICER_FRIGATE_GROUPS,
    RAT_OFFICER_GROUPS,
    RAT_SUPER_GROUPS,
    RAT_TITAN_GROUPS,
)

from .. import app_settings, models

logger = get_extension_logger(__name__)


class Paginator(LimitOffsetPagination):
    class Input(Schema):
        limit: int = Field(app_settings.CT_PAGINATION_SIZE, ge=1)
        offset: int = Field(0, ge=0)

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        **params: DictStrAny,
    ) -> any:
        offset = pagination.offset
        limit: int = pagination.limit
        return {
            "items": queryset[offset: offset + limit],
            "count": self._items_count(queryset),
        }


def get_main_character(request, character_id):
    perms = True
    main_char = EveCharacter.objects\
        .select_related('character_ownership', 'character_ownership__user__profile', 'character_ownership__user__profile__main_character', )\
        .get(character_id=character_id)
    try:
        main_char = main_char.character_ownership.user.profile.main_character
    except (ObjectDoesNotExist):
        pass

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(
        request.user)
    if main_char not in visible:
        account_chars = request.user.profile.main_character.character_ownership.user.character_ownerships.all(
        )
        logger.warning(
            f"{request.user} Can See {list(visible)}, requested {main_char.id}")
        if main_char in account_chars:
            pass
        else:
            perms = False

    if not request.user.has_perm('corptools.view_characteraudit'):
        logger.warning(
            f"{request.user} does not have Perm requested, Requested {main_char.id}")
        perms = False

    return perms, main_char


def get_alts_queryset(main_char):
    try:
        linked_characters = main_char.character_ownership.user.character_ownerships.all(
        ).values_list('character_id', flat=True)

        return EveCharacter.objects.filter(id__in=linked_characters)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def get_corporation_characters(request, corporation_id):
    return EveCharacter.objects.filter(
        character_ownership__user__profile__main_character__corporation_id=corporation_id,
        character_id__in=models.CharacterAudit.objects.visible_eve_characters(
            request.user
        ).values_list("character_id", flat=True)
    )


def round_or_null(value, digits=2):
    if value:
        return round(value, digits)
    else:
        return value


def wallet_check(characters, types, first_parties=None, minimum_amount=None, look_back=30):
    start_date = timezone.now() - timedelta(days=look_back)
    if isinstance(types, str):
        types = [types]

    if isinstance(first_parties, str):
        first_parties = [first_parties]

    qry = models.CharacterWalletJournalEntry.objects.filter(
        character__character__in=characters,
        ref_type__in=types,
        date__gte=start_date)

    if first_parties:
        qry = qry.filter(
            first_party_id__in=first_parties
        )

    if minimum_amount:
        qry = qry.filter(
            amount__gte=minimum_amount
        )

    return qry


def mining_check(characters, groups, look_back=30):
    start_date = timezone.now() - timedelta(days=look_back)
    return models.CharacterMiningLedger.objects.filter(
        character__character__in=characters,
        type_name__group_id__in=groups,
        date__gte=start_date
    )


def bounty_check(characters, groups, look_back=30):
    start_date = timezone.now() - timedelta(days=look_back)
    return models.CharacterBountyStat.objects.filter(
        event__entry__character__character__in=characters,
        type_name__group_id__in=groups,
        event__entry__date__gte=start_date
    )


def rats_killed_check(characters, groups):
    return models.CharacterBountyStat.objects.filter(
        event__entry__character__character__in=characters,
        type_name__group_id__in=groups
    )


def glance_rest_count(characters):
    excluded_groups = RAT_CAPITAL_GROUPS
    excluded_groups += RAT_SUPER_GROUPS
    excluded_groups += RAT_TITAN_GROUPS
    excluded_groups += RAT_OFFICER_GROUPS
    excluded_groups += RAT_OFFICER_CRUISER_GROUPS
    excluded_groups += RAT_OFFICER_FRIGATE_GROUPS

    return models.CharacterBountyStat.objects.filter(
        event__entry__character__character__in=characters
    ).exclude(
        type_name__group_id__in=excluded_groups
    ).aggregate(total=Sum("qty"))["total"]


def glance_capital_count(characters):
    return rats_killed_check(
        characters,
        RAT_CAPITAL_GROUPS
    ).aggregate(total=Sum("qty"))["total"]


def glance_supers_count(characters):
    return rats_killed_check(
        characters,
        RAT_SUPER_GROUPS
    ).aggregate(total=Sum("qty"))["total"]


def glance_titans_count(characters):
    return rats_killed_check(
        characters,
        RAT_TITAN_GROUPS
    ).aggregate(total=Sum("qty"))["total"]


def glance_officers_count(characters):
    return rats_killed_check(
        characters,
        RAT_OFFICER_GROUPS
    ).aggregate(total=Sum("qty"))["total"]


def glance_officers_cruiser_count(characters):
    return rats_killed_check(
        characters,
        RAT_OFFICER_CRUISER_GROUPS
    ).aggregate(total=Sum("qty"))["total"]


def glance_officers_frigate_count(characters):
    return rats_killed_check(
        characters,
        RAT_OFFICER_FRIGATE_GROUPS
    ).aggregate(total=Sum("qty"))["total"]


def glance_incursion_check(characters):
    from_ids = [1000125]
    types = ["corporate_reward_payout"]
    return wallet_check(characters, types, first_parties=from_ids).aggregate(total=Sum("amount"))["total"]


def glances_missions_check(characters):
    types = ["agent_mission_reward", "agent_mission_time_bonus_reward"]
    return wallet_check(characters, types).aggregate(total=Sum("amount"))["total"]


def glances_ratting_check(characters):
    types = ["bounty_prizes"]
    min_amount = 1000000
    # 5 mill ticks should cover gate rats etc. but still show passive ratting
    return wallet_check(characters, types, minimum_amount=min_amount).aggregate(total=Sum("amount"))["total"]
    # return wallet_check(characters, types).aggregate(total=Sum("amount"))["total"]


def glances_pochven_check(characters):
    from_ids = [1000298]
    types = ["corporate_reward_payout"]
    return wallet_check(characters, types, first_parties=from_ids).aggregate(total=Sum("amount"))["total"]


def glances_market_check(characters):
    types = ["brokers_fee", "market_provider_tax",
             "market_transaction", "transaction_tax"]
    return wallet_check(characters, types).aggregate(total=Sum("amount"))["total"]


def glances_industry_check(characters):
    types = ["industry_job_tax"]
    return wallet_check(characters, types).count()


def glances_pi_check(characters):
    types = ["planetary_import_tax",
             "planetary_export_tax"]
    return wallet_check(characters, types).aggregate(total=Sum("amount"))["total"]


def glances_ore_check(characters):
    return mining_check(
        characters,
        MINING_ORE_GROUPS
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def glances_moon_check(characters):
    return mining_check(
        characters,
        MINING_MOON_GROUPS
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def glances_ice_check(characters):
    return mining_check(
        characters,
        MINING_ICE_GROUPS
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def glances_gas_check(characters):
    return mining_check(
        characters,
        MINING_GAS_GROUPS
    ).aggregate(
        total=Sum(F("quantity") * F("type_name__volume"))
    )["total"]


def roundFloat(input):
    if input:
        return int(input)
    else:
        return input


def glances_assets_character(characters):
    cats = [6, 65, 22]
    injector_grp = 1739
    sp_types = [
        40519,  # extractor
    ]

    ship_assets = models.CharacterAsset.objects.filter(
        Q(type_name__group__category_id__in=cats)
        | Q(type_name__group_id=injector_grp),
        character__character__in=characters,
    ).values(
        'type_name__group__id'
    ).annotate(
        grp_total=Sum('quantity')
    ).annotate(
        grp_name=F('type_name__group__name')
    ).annotate(
        grp_id=F('type_name__group_id')
    ).annotate(
        cat_id=F('type_name__group__category_id')
    ).order_by(
        '-grp_total'
    )

    sp_assets = models.CharacterAsset.objects.filter(
        character__character__in=characters,
        type_name__id__in=sp_types
    ).values(
        'type_name__id'
    ).annotate(
        type_total=Sum('quantity')
    )

    den_assets = models.CharacterAsset.objects.filter(
        character__character__in=characters,
        type_name__id=85230,
        location_type="solar_system"
    ).values(
        'type_name__id'
    ).annotate(
        type_total=Sum('quantity')
    )

    sp_assets = sp_assets | den_assets

    return assets_glances(ship_assets, sp_assets)


def glances_assets_corporation(characters, corp_id, user=None):
    cats = [6, 65]
    injector_grp = 1739
    sp_types = [
        40519,  # extractor
    ]

    ship_assets = models.CorpAsset.objects.get_queryset()
    if user:
        ship_assets = models.CorpAsset.get_visible(
            user
        )

    ship_assets = models.CorpAsset.objects.filter(
        Q(type_name__group__category_id__in=cats)
        | Q(type_name__group_id=injector_grp),
        corporation__corporation__corporation_id=corp_id,
    ).values(
        'type_name__group__id'
    ).annotate(
        grp_total=Sum('quantity')
    ).annotate(
        grp_name=F('type_name__group__name')
    ).annotate(
        grp_id=F('type_name__group_id')
    ).annotate(
        cat_id=F('type_name__group__category_id')
    ).order_by(
        '-grp_total'
    )

    sp_assets = models.CorpAsset.objects.get_queryset()

    if user:
        sp_assets = models.CorpAsset.get_visible(
            user
        )

    sp_assets = models.CorpAsset.objects.filter(
        corporation__corporation__corporation_id=corp_id,
        type_name__id__in=sp_types
    ).values(
        'type_name__id'
    ).annotate(
        type_total=Sum('quantity')
    )

    return {
        "corporate": assets_glances(ship_assets, sp_assets),
        "character": glances_assets_character(characters)
    }


def assets_glances(ship_assets, sp_assets):
    injector_grp = 1739
    mining_groups = [463, 543]
    frig_groups = [25, 324, 830, 831, 834, 893, 1283, 1527, 2078]
    destroyer_groups = [420, 541, 1305, 1534]
    cruiser_groups = [26, 358, 832, 833, 894, 906, 963, 1972]
    bc_groups = [419, 540, 1201]
    bs_groups = [27, 381, 898, 900]
    indy_groups = [28, 380, 1202]
    indy_command_groups = [941]
    dread_groups = [485, 4594]
    cap_indy_groups = [902, 513, 883]
    citadel_groups = [1657,]
    eng_comp_groups = [1404, ]
    refinary_groups = [1406, 4744]
    flex_groups = [1408, 2016, 2017]
    merc_den_groups = [4810, ]

    out_groups = {
        "frigate": 0,
        "destroyer": 0,
        "cruiser": 0,
        "battlecruiser": 0,
        "battleship": 0,
        "carrier": 0,
        "fax": 0,
        "dread": 0,
        "supercarrier": 0,
        "titan": 0,
        "mining": 0,
        "hauler": 0,
        "indy_command": 0,
        "capital_indy": 0,
        "injector": 0,
        "extractor": 0,
        "citadel": 0,
        "eng_comp": 0,
        "refinary": 0,
        "flex": 0,
        "merc_den_grp": 0,
        "merc_den": 0,
    }

    for group in ship_assets:
        # this is ugly... but functional...
        grp = group["type_name__group__id"]

        if grp in frig_groups:
            out_groups["frigate"] += group["grp_total"]
        elif grp in destroyer_groups:
            out_groups["destroyer"] += group["grp_total"]
        elif grp in cruiser_groups:
            out_groups["cruiser"] += group["grp_total"]
        elif grp in bc_groups:
            out_groups["battlecruiser"] += group["grp_total"]
        elif grp in bs_groups:
            out_groups["battleship"] += group["grp_total"]
        elif grp in indy_groups:
            out_groups["hauler"] += group["grp_total"]
        elif grp in indy_command_groups:
            out_groups["indy_command"] += group["grp_total"]
        elif grp in mining_groups:
            out_groups["mining"] += group["grp_total"]
        elif grp in dread_groups:
            out_groups["dread"] += group["grp_total"]
        elif grp in cap_indy_groups:
            out_groups["capital_indy"] += group["grp_total"]
        elif grp == 30:
            out_groups["titan"] += group["grp_total"]
        elif grp == 547:
            out_groups["carrier"] += group["grp_total"]
        elif grp == 1538:
            out_groups["fax"] += group["grp_total"]
        elif grp == 659:
            out_groups["supercarrier"] += group["grp_total"]
        elif grp == injector_grp:
            out_groups["injector"] += group["grp_total"]
        elif grp in citadel_groups:
            out_groups["citadel"] += group["grp_total"]
        elif grp in eng_comp_groups:
            out_groups["eng_comp"] += group["grp_total"]
        elif grp in refinary_groups:
            out_groups["refinary"] += group["grp_total"]
        elif grp in flex_groups:
            out_groups["flex"] += group["grp_total"]
        elif grp in merc_den_groups:
            out_groups["merc_den_grp"] += group["grp_total"]

    for sp_type in sp_assets:
        _type = sp_type["type_name__id"]
        if _type == 40519:
            out_groups["extractor"] += sp_type["type_total"]
        elif _type == 85230:
            out_groups["merc_den"] += sp_type["type_total"]

    return out_groups



# Standard Library
from datetime import timedelta
from unittest.mock import Mock, patch

# Django
from django.test import TestCase
from django.utils import timezone

# AA Example App
from corptools.models import EveName
from corptools.tasks.updates import update_eve_name


class TestUpdateEveName(TestCase):

    def setUp(self):
        """
        Setup test data

        :return:
        :rtype:
        """
        self.character_id = 12345
        self.corporation_id = 67890
        self.alliance_id = 11111

        self.character = EveName.objects.create(
            eve_id=self.character_id,
            name="Test Character",
            category=EveName.CHARACTER,
        )

        self.corporation = EveName.objects.create(
            eve_id=self.corporation_id,
            name="Test Corporation",
            category=EveName.CORPORATION,
        )

        self.alliance = EveName.objects.create(
            eve_id=self.alliance_id,
            name="Test Alliance",
            category=EveName.ALLIANCE,
        )

        past = timezone.now() - timedelta(days=31)
        EveName.objects.filter(
            eve_id=self.character_id).update(last_update=past)
        EveName.objects.filter(
            eve_id=self.corporation_id).update(last_update=past)
        EveName.objects.filter(
            eve_id=self.alliance_id).update(last_update=past)

    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_retries_when_error_flag_is_set(self, mock_get_error_flag):
        """
        Test retries when error flag is set

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :return:
        :rtype:
        """
        mock_get_error_flag.return_value = True

        with patch.object(
            update_eve_name, "retry", side_effect=Exception("Retry called")
        ):
            with self.assertRaises(Exception) as context:
                update_eve_name(self.character_id)

            self.assertEqual(str(context.exception), "Retry called")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_does_not_update_when_name_does_not_need_update(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test does not update when not needed

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        self.character.last_update = timezone.now()
        self.character.save()

        update_eve_name(self.character_id)

        mock_eve_names.get_character.assert_not_called()

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_character_name_with_no_corporation_or_alliance(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates cvharacter information, but ignores alliance and corp info

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_update = Mock()
        mock_update.name = "Updated Character Name"
        mock_update.alliance = None
        mock_update.corp = None
        mock_eve_names.get_character.return_value = mock_update

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        self.assertEqual(self.character.name, "Updated Character Name")
        self.assertIsNone(self.character.alliance)
        self.assertIsNone(self.character.corporation)

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_character_with_alliance_and_corporation(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates character information with corp and alliance

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_alliance = Mock()
        mock_alliance.id = 99999
        mock_alliance.name = "New Alliance"

        mock_corp = Mock()
        mock_corp.id = 88888
        mock_corp.name = "New Corporation"

        mock_update = Mock()
        mock_update.name = "Character With Corp And Alliance"
        mock_update.alliance = mock_alliance
        mock_update.corp = mock_corp
        mock_eve_names.get_character.return_value = mock_update

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        self.assertEqual(self.character.name,
                         "Character With Corp And Alliance")
        self.assertIsNotNone(self.character.alliance)
        self.assertEqual(self.character.alliance.eve_id, 99999)
        self.assertEqual(self.character.alliance.name, "New Alliance")
        self.assertIsNotNone(self.character.corporation)
        self.assertEqual(self.character.corporation.eve_id, 88888)
        self.assertEqual(self.character.corporation.name, "New Corporation")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_sets_corporation_alliance_when_character_has_alliance(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test sets corp and alliance when character has alliance

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_alliance = Mock()
        mock_alliance.id = 55555
        mock_alliance.name = "Corporation Alliance"

        mock_corp = Mock()
        mock_corp.id = 44444
        mock_corp.name = "Corporation With Alliance"

        mock_update = Mock()
        mock_update.name = "Character Name"
        mock_update.alliance = mock_alliance
        mock_update.corp = mock_corp
        mock_eve_names.get_character.return_value = mock_update

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        corporation = EveName.objects.get(eve_id=44444)
        self.assertEqual(corporation.alliance_id, 55555)

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_corporation_info(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates corporation info

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_update = Mock()
        mock_update.corporation_name = "Updated Corporation Name"
        mock_update.alliance = None
        mock_eve_names.get_corp.return_value = mock_update

        update_eve_name(self.corporation_id)

        self.corporation.refresh_from_db()
        self.assertEqual(self.corporation.name, "Updated Corporation Name")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_corporation_with_alliance(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates corporation with alliance

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_alliance = Mock()
        mock_alliance.id = 77777
        mock_alliance.name = "Corporation's Alliance"

        mock_update = Mock()
        mock_update.corporation_name = "Corporation With Alliance Name"
        mock_update.alliance = mock_alliance
        mock_eve_names.get_corp.return_value = mock_update

        update_eve_name(self.corporation_id)

        self.corporation.refresh_from_db()
        self.assertEqual(self.corporation.name,
                         "Corporation With Alliance Name")
        self.assertIsNotNone(self.corporation.alliance)
        self.assertEqual(self.corporation.alliance.eve_id, 77777)
        self.assertEqual(self.corporation.alliance.name,
                         "Corporation's Alliance")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_alliance_info(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates alliance info

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_update = Mock()
        mock_update.name = "Updated Alliance Name"
        mock_eve_names.get_alliance.return_value = mock_update

        update_eve_name(self.alliance_id)

        self.alliance.refresh_from_db()
        self.assertEqual(self.alliance.name, "Updated Alliance Name")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_sets_error_flag_when_exception_has_low_error_limit(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test sets error flag when exception has low error limit

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        exception = Exception("ESI Error")
        exception.response = Mock()
        exception.response.headers = {"x-esi-error-limit-remain": "10"}
        mock_eve_names.get_character.side_effect = exception

        update_eve_name(self.character_id)

        mock_set_error.assert_called_once()

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_does_not_set_error_flag_when_error_limit_is_high(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test does not set error flag when error limit is high

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        exception = Exception("ESI Error")
        exception.response = Mock()
        exception.response.headers = {"x-esi-error-limit-remain": "100"}
        mock_eve_names.get_character.side_effect = exception

        update_eve_name(self.character_id)

        mock_set_error.assert_not_called()

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_last_update_when_exception_occurs(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates last_update when exception occurs

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """
        mock_get_error_flag.return_value = False

        exception = Exception("ESI Error")
        mock_eve_names.get_character.side_effect = exception

        old_last_update = self.character.last_update

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        self.assertGreater(self.character.last_update, old_last_update)

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_handles_exception_without_response_attribute(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test handles exception without response attribute

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        exception = Exception("Generic Error")
        mock_eve_names.get_character.side_effect = exception

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        mock_set_error.assert_not_called()

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_handles_exception_with_response_but_missing_header(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test handles exception without response attribute

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        exception = Exception("Error without specific header")
        exception.response = object()
        mock_eve_names.get_character.side_effect = exception

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        mock_set_error.assert_not_called()

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_creates_new_alliance_when_updating_character(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test creates new alliance when updating character

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        new_alliance_id = 123456

        mock_alliance = Mock()
        mock_alliance.id = new_alliance_id
        mock_alliance.name = "Brand New Alliance"

        mock_update = Mock()
        mock_update.name = "Character Name"
        mock_update.alliance = mock_alliance
        mock_update.corp = None
        mock_eve_names.get_character.return_value = mock_update

        self.assertFalse(EveName.objects.filter(
            eve_id=new_alliance_id).exists())

        update_eve_name(self.character_id)

        self.assertTrue(EveName.objects.filter(
            eve_id=new_alliance_id).exists())
        alliance = EveName.objects.get(eve_id=new_alliance_id)
        self.assertEqual(alliance.name, "Brand New Alliance")
        self.assertEqual(alliance.category, EveName.ALLIANCE)

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_creates_new_corporation_when_updating_character(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test creates new corporation when updating character

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        new_corp_id = 654321

        mock_corp = Mock()
        mock_corp.id = new_corp_id
        mock_corp.name = "Brand New Corporation"

        mock_update = Mock()
        mock_update.name = "Character Name"
        mock_update.alliance = None
        mock_update.corp = mock_corp
        mock_eve_names.get_character.return_value = mock_update

        self.assertFalse(EveName.objects.filter(eve_id=new_corp_id).exists())

        update_eve_name(self.character_id)

        self.assertTrue(EveName.objects.filter(eve_id=new_corp_id).exists())
        corporation = EveName.objects.get(eve_id=new_corp_id)
        self.assertEqual(corporation.name, "Brand New Corporation")
        self.assertEqual(corporation.category, EveName.CORPORATION)

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_existing_alliance(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates existing alliance

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        existing_alliance = EveName.objects.create(
            eve_id=999888, name="Old Alliance Name", category=EveName.ALLIANCE
        )

        mock_alliance = Mock()
        mock_alliance.id = 999888
        mock_alliance.name = "Updated Alliance Name"

        mock_update = Mock()
        mock_update.corporation_name = "Corp Name"
        mock_update.alliance = mock_alliance
        mock_eve_names.get_corp.return_value = mock_update

        update_eve_name(self.corporation_id)

        existing_alliance.refresh_from_db()
        self.assertEqual(existing_alliance.name, "Updated Alliance Name")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_updates_existing_corporation_when(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test updates existing corporation

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        existing_corp = EveName.objects.create(
            eve_id=888999, name="Old Corp Name", category=EveName.CORPORATION
        )

        mock_corp = Mock()
        mock_corp.id = 888999
        mock_corp.name = "Updated Corp Name"

        mock_update = Mock()
        mock_update.name = "Character Name"
        mock_update.alliance = None
        mock_update.corp = mock_corp
        mock_eve_names.get_character.return_value = mock_update

        update_eve_name(self.character_id)

        existing_corp.refresh_from_db()
        self.assertEqual(existing_corp.name, "Updated Corp Name")

    @patch("corptools.tasks.updates.set_error_count_flag")
    @patch("corptools.tasks.updates.eve_names")
    @patch("corptools.tasks.updates.get_error_count_flag")
    def test_handles_character_with_corporation_but_no_alliance(
        self, mock_get_error_flag, mock_eve_names, mock_set_error
    ):
        """
        Test handles character with corporation but no alliance

        :param mock_get_error_flag:
        :type mock_get_error_flag:
        :param mock_eve_names:
        :type mock_eve_names:
        :param mock_set_error:
        :type mock_set_error:
        :return:
        :rtype:
        """

        mock_get_error_flag.return_value = False

        mock_corp = Mock()
        mock_corp.id = 111222
        mock_corp.name = "Corp Without Alliance"

        mock_update = Mock()
        mock_update.name = "Character Name"
        mock_update.alliance = None
        mock_update.corp = mock_corp
        mock_eve_names.get_character.return_value = mock_update

        update_eve_name(self.character_id)

        self.character.refresh_from_db()
        corporation = EveName.objects.get(eve_id=111222)
        self.assertIsNone(corporation.alliance_id)
